.column-center
{
    text-align: center;
}

/* child row selection management */
thead .details-control,
td.details-control {
    text-align: center;
    color: forestgreen;
    cursor: pointer;
}
thead.allshown th.details-control,
tr.shown td.details-control {
    color: deepskyblue;
}
thead.someshown th.details-control {
    color:grey;
}

/* edit selection/control management */
td.edit-control {
    text-align: center;
    color: forestgreen;
    cursor: pointer;
}
tr.editing td.edit-control {
    color: deepskyblue;
}
tr.needsedit td.edit-control {
    color:orangered;
}

/* datatables Editor */
div.DTE_Field input[type="checkbox"] {
    vertical-align: middle;
}

/* childrow display */

/* indent child rows */
.childrow-display {
    margin-left: 20px;
    width: 95%;
}

/* from https://datatables.net/blog/2017-06-30#Example */

.childrow-display .DTE_Label {
    font-weight: bold;
    padding-top: 6px;
}

.childrow-editform div.DTE div.DTE_Body div.DTE_Body_Content div.DTE_Field,
.childrow-display p.DTE_Field_Input
{
    padding: 0.7em 0 0 0;
}

.childrow-display .DTE_Label,
.childrow-editform div.DTE div.DTE_Body div.DTE_Body_Content div.DTE_Field > label {
    float: none;
    clear: both;
    width: 100%;
    font-size: 1.3em;
}
.childrow-editform div.DTE div.DTE_Body div.DTE_Body_Content div.DTE_Field > div.DTE_Field_Input {
    float: none;
    clear: both;
    width: 100%;
    padding-top: 1px;
}

.childrow-editform div.DTE div.DTE_Header,
.childrow-editform div.DTE div.DTE_Footer {
    background-color: transparent;
    border-color: black;
    display: block;
    float: left;
    padding-top: 6px;
}

/* override editor.dataTables.css */
.childrow-editform div.DTE div.DTE_Form_Buttons button {
    float:left;
}

.childrow-editform div.DTE div.DTE_Header {
    /*height: 60px;*/
    font-size: 2em;
}

/* overrides padding from editor.jqueryui.css */
.childrow-editform div.DTE div.DTE_Body {
    padding: 52px 0;
}

.childrow-editform p.start-editing,
.childrow-editform p.add-new {
    text-align: center;
    font-size: 1.4em;
    line-height: 1.4em;
}

.childrow-editform p.start-editing {
    padding-top: 7em;
}

.childrow-display .DTE_Field_Input {
    padding: 0 6px;
}

/* display border around child row when shown */
table.dataTable {
    /* override dataTables.jqueryui, required for tr border */
    border-collapse: collapse;
}
tr.shown {
    border-top: 1px solid black;
    border-left: 1px solid black;
    border-right: 1px solid black;
}
tr.shown + tr {
    border-left: 1px solid black;
    border-right: 1px solid black;
    border-bottom: 1px solid black;
}

/* shrink to fit, see hack at https://stackoverflow.com/a/43615091/799921 */
.shrink-to-fit {
    width:0.1%;
    white-space: nowrap;
}

/* ckeditor customization */
.ck.ck-editor__editable_inline {
    border-width: 2px !important;
    border-style: inset !important;
    border-color: #767676 #c3c3c3 #c3c3c3 #767676 !important;
    background-color: white;
}

/* select items displayed as tags, matches .select2-container--default .select2-selection--multiple .select2-selection__choice */
/* see datatables.render_select_as_tags() */
.tag-display {
    background-color: #e4e4e4;
    border: 1px solid #aaa;
    border-radius: 4px;
    cursor: default;
    float: left;
    margin-right: 5px;
    margin-top: 5px;
    padding: 0 5px;
    white-space: nowrap;    /* added beyond .select2 styling */
}